//#include "stdio.h"
//#include <stdio.h>
#include "WProgram.h"
#include "XBeeLibrary.h"
#include <float.h>
#include <limits.h>



// string
const prog_char string_reading_len1[] PROGMEM = "V001\r\n";  
const prog_char string_ignore_byte[] PROGMEM  = "V002\r\n";  
const prog_char string_reading_len2[] PROGMEM  = "V003\r\n";  
const prog_char string_reading_api_id_value[] PROGMEM  = "V004\r\n";  
const prog_char string_read_temperature[] PROGMEM  = "V005\r\n";  
const prog_char string_read_heading[] PROGMEM  = "V006\r\n";  
const prog_char string_sending[] PROGMEM  = "V007\r\n";  
const prog_char string_byte_read[] PROGMEM  = "V008\r\n";  
const prog_char string_message_received[] PROGMEM  = "V009\r\n";  
const prog_char string_frame_control_byte[] PROGMEM  = "V010\r\n";  
const prog_char string_transaction_sequence[] PROGMEM  = "V011\r\n";  
const prog_char string_command_byte[] PROGMEM  = "V012\r\n";
const prog_char string_reading_cmd_data[] PROGMEM  = "V013\r\n";
const prog_char string_reading_checksum[] PROGMEM  = "V014\r\n";
const prog_char string_checksum_ok[] PROGMEM  = "V015\r\n";
const prog_char string_unknown_state[] PROGMEM  = "V016\r\n";
const prog_char string_explicit_rx_ind[] PROGMEM  = "V017\r\n";
const prog_char string_general_command[] PROGMEM  = "V018\r\n";
const prog_char string_non_general_command[] PROGMEM  = "V019\r\n";
const prog_char string_non_manufacturer_specific[] PROGMEM  = "V020\r\n";
const prog_char string_manufacturer_specific[] PROGMEM  = "V021\r\n";
const prog_char string_client_to_server[] PROGMEM  = "V022\r\n";
const prog_char string_server_to_client[] PROGMEM  = "V023\r\n";
const prog_char string_read_command[] PROGMEM  = "V024\r\n";
const prog_char string_unknown_attr[] PROGMEM  = "V025\r\n";
const prog_char string_unknown_command_identifier[] PROGMEM  = "V026\r\n";
const prog_char string_send_response[] PROGMEM  = "V027\r\n";
const prog_char string_not_implemented[] PROGMEM  = "V028\r\n";
const prog_char string_unknown_clusterid[] PROGMEM  = "V029\r\n";
const prog_char string_read_illuminance[] PROGMEM  = "V030\r\n";
const prog_char string_checksum_incorrect[] PROGMEM  ="V031\r\n";
//const prog_char string_not_implemented[] PROGMEM  = "V031\r\n";
//const prog_char string_read_illuminance[] PROGMEM  = "read illuminance\r\n";
//const prog_char string_server_to_client[] PROGMEM  = "server to client\r\n";
//const prog_char string_more_90[] PROGMEM  = "more 90\r\n";




XBee::XBee(struct zigbee *zb)
{
  _state = WAITING_FOR_NEW_PACKET;
  _zb = zb;
  _frameId = 0;
  Serial.begin(9600);
}

void XBee::serve()
{
  while (Serial.available() > 0) {
    byte incomingByte = Serial.read();
    printPgm(string_byte_read);
    printHex(&incomingByte, 1);
    switch(_state) {
      case WAITING_FOR_NEW_PACKET:
        if (0x7E == incomingByte) {
          _state = READING_LEN1;
          printPgm(string_reading_len1);
        } else {
          printPgm(string_ignore_byte);          
        }
        break;
      case READING_LEN1:
        _msgLength = (incomingByte << 8) & 0xFF00;
        _state = READING_LEN2; 
        printPgm(string_reading_len2);          
        
        break;
      case READING_LEN2:
        _msgLength = _msgLength | (incomingByte & 0x00FF);       
        _state = READING_API_ID_VALUE; 
        printPgm(string_reading_api_id_value);
        break;        
      case READING_API_ID_VALUE:
        _apiIdValue = incomingByte;
        _msgBufferIndex = 0;
        _state = READING_CMD_DATA;
        printPgm(string_reading_cmd_data);
        break;
        
      case READING_CMD_DATA:
        _msgBuffer[_msgBufferIndex] = incomingByte;
        _msgBufferIndex = _msgBufferIndex + 1; 
        if (_msgBufferIndex > (_msgLength-2)) { // -2 buffer start 1 byte after apiValue (-1) and ends before checksum (-1)
          _state = READING_CHECKSUM;
          printPgm(string_reading_checksum);
        }
        break;
      case READING_CHECKSUM:
        {
        printPgm(string_message_received);
        
        //printHex(_msgBuffer,_msgBufferIndex);
        int checksum_add = _apiIdValue;
        //checksum_add += _frameId;    
        for(int i=0;i<_msgBufferIndex;i++) {
          checksum_add += _msgBuffer[i];
        } 
        byte checksum = 0xFF -(checksum_add & 0xFF);
        //printPgm(string_message_received);
        //printHex(&checksum,1);
        //printHex(&incomingByte,1);
        
        if (checksum != incomingByte) {
          //errorPgm("checksum incorrect\r\n");
           errorPgm(string_checksum_incorrect);
        
        } else {
          printPgm(string_checksum_ok);
          parse();
        }
        _state = WAITING_FOR_NEW_PACKET;
        }
        break;
      default:
        errorPgm(string_unknown_state);
        break;
    }
    if (_zb->fprogress != NULL) _zb->fprogress(); // show progress tipically blink the led   
  }
  
  return;
}

void XBee::parse()
{
  
  switch(_apiIdValue) 
  {
    case EXPLICIT_RX_IND:
      //errorPgm(string_explicit_rx_ind);
      zb_explicit_rx_packet();
      break;
    case ZB_RX_PACKET:
      zb_rx_packet();
      break;
    case MODEM_STATUS:
      modemStatus(_msgBuffer[0]);
      break;
    case AT_CMD_RES:
      errorPgm(string_not_implemented);
      break;
    case REMOTE_CMD_RES:
      errorPgm(string_not_implemented);
      break;
    case ZB_TX_STATUS:
      errorPgm(string_not_implemented);
      break; 
    case ZB_IO_DATA_SAMPLE_RX_IND:
      errorPgm(string_not_implemented);
      break;
    case XB_SENSOR_READ_IND:
      errorPgm(string_not_implemented);
      break;
    case NODE_ID_IND:
      errorPgm(string_not_implemented);
      break;
    default:
      errorPgm(string_not_implemented);
      break;
  }
}


void xbee_init(struct zigbee *zb) 
{
  memset(zb, 0, sizeof(struct zigbee));
}

void XBee::error(char *msg) 
{
        if (_zb->ferror != NULL) _zb->ferror(msg);  
}

void XBee::errorPgm(PGM_P str) {
  char buffer[30];
  strcpy_P(buffer, str);
  error(buffer);
}


void XBee::printPgm(PGM_P str) {
  char buffer[30];
  strcpy_P(buffer, str);
  print(buffer);
}


void XBee::print(char *msg) 
{
  if (_zb->fprint != NULL) {
    _zb->fprint(msg);
  }  
}

void XBee::printHex(byte *msg,int size)
{
    if (_zb->fprintHex != NULL) _zb->fprintHex(msg, size);
}

void XBee::modemStatus(int status) 
{
    if (_zb->fmodemStatus != NULL) _zb->fmodemStatus(status);
}

float XBee::getTemperature() 
{
    if (_zb->fgetTemperature != NULL) return _zb->fgetTemperature();
    return FLT_MAX; //TODO:float max
}
    
float XBee::getHeading() 
{
    if (_zb->fgetHeading != NULL) return _zb->fgetHeading();
    return FLT_MAX; //TODO:float max
}

float XBee::getTilt(int i) 
{
    if (_zb->fgetTilt != NULL) return _zb->fgetTilt(i);
    return FLT_MAX; //TODO:float max
}
    


void XBee::zb_rx_packet()
{
   //TODO:print("RX_IND AO=0\r\n");
   memcpy(&_sourceAddress64, _msgBuffer, 8);
   memcpy(&_sourceAddress16, _msgBuffer+8, 2);
   //byte 10 is packet acknowledge 01 packet acknowledged
  //TODO: print ack 
   parse_rx_packet_payload(11);
}

void XBee::zb_explicit_rx_packet()
{
   printPgm(string_explicit_rx_ind);
   memcpy(&_sourceAddress64, _msgBuffer, 8);
   memcpy(&_sourceAddress16, _msgBuffer+8, 2);
   byte source_endpoint = _msgBuffer[10];
   byte dest_endpoint = _msgBuffer[11];
   _clusterid = _msgBuffer[12]<<8;
   _clusterid = _clusterid + _msgBuffer[13];
   int profileid = (_msgBuffer[14]<<8) | _msgBuffer[15];
   byte options = _msgBuffer[16];
   //byte 16 is packet acknowledge 01 packet acknowledged
  //TODO: print ack 
   parse_rx_packet_payload(17);
}



void XBee::parse_rx_packet_payload(int i) 
{
   printPgm(string_frame_control_byte);
   //printHex(_msgBuffer+i,1);
   if ((_msgBuffer[i] & 0xC0) == 0) // 0xC0 = 0b1100 0000 
   {
     printPgm(string_general_command); 
     parse_general_command(i);      
   } else {
     errorPgm(string_non_general_command);
   }
  
}


void XBee::parse_general_command(int i)
{
     //frame control field bit 0 is the leftmost bit
     //bit 0-1 frame type
     //bit 2 manufacturer specific
     //bit 3 direction ( 0 client to server, 1 server to client)
     //bit 4  disable default response
     //bit 5-7 reserved
     
     boolean manSpecific = (_msgBuffer[i] & 0x20) == 0;
     int offset = 0;
     if (manSpecific) { //frame control field, manufacturer specific 
       printPgm(string_non_manufacturer_specific);
       
     } else {
       printPgm(string_manufacturer_specific);
       offset = 2;
     }  
       
       if ((_msgBuffer[i] & 0x10) == 0) {
         printPgm(string_client_to_server);
         sendDefaultResponseIfNeeded(i);
         _transaction_sequence = _msgBuffer[i+1+offset];
         printPgm(string_transaction_sequence);
         //printHex(&_transaction_sequence,1);
         parse_command(i+2+offset, manSpecific);
  
       } else {
         errorPgm(string_server_to_client);
       }
}

void XBee::sendDefaultResponseIfNeeded(int i)
{
  if ((_msgBuffer[i] & 0x08) != 0) {
    errorPgm(string_not_implemented);
    //todo: send default response
    //TODO:print("enable default response\r\n");
  }
}


void XBee::parse_command(int i, boolean isManSpecific)
{
       // command 
       printPgm(string_command_byte);
       //printInt(i);
       //print("\r\n");
       //printHex(_msgBuffer+i,1);
       switch(_msgBuffer[i]) 
       {
         case READ_ATTRS:
           printPgm(string_read_command);
           //TODO read several attr values at once
           
           switch(_clusterid) {
             case TEMP_MEASUREMENT_CLUSTERID:
               read_temp_measurement_cluster(i);
               break;
             case ILLUMINANCE_MEASUREMENT_CLUSTERID:
               read_illuminance_measurement_cluster(i);
               break;
             case TILT_MEASUREMENT_CLUSTERID:
               read_tilt_measurement_cluster(i);
               break;
             case HEADING_MEASUREMENT_CLUSTERID:
               read_heading_measurement_cluster(i);
               break;
               
             default:
               errorPgm(string_unknown_clusterid);
            }           
           break;
         default:
           //TODO: send UNSUP_GENERAL COMMAND
           errorPgm(string_unknown_command_identifier);
           break;
       }  
}

void XBee::read_temp_measurement_cluster(int i)
{
  int attr = _msgBuffer[i+1] << 8;
  attr = attr + _msgBuffer[i+2];
  attr = attr & 0xFFFF;
  //printHex(_msgBuffer+i+1,2); 

  switch(attr)
  {
    case TEMP_MEASURED_VALUE:
      read_temp();                
      break;
    case TEMP_MIN_MEASURED_VALUE:
      read_temp_min();                
      break;
    case TEMP_MAX_MEASURED_VALUE:
      read_temp_max();                
      break;
    case TEMP_TOLERANCE:
      read_temp_tolerance();                
      break;               
    default:
      errorPgm(string_unknown_attr);
      break;
  }  
}

void XBee::read_heading_measurement_cluster(int i)
{
  int attr = _msgBuffer[i+1] << 8;
  attr = attr + _msgBuffer[i+2];
  attr = attr & 0xFFFF;
  //printHex(_msgBuffer+i+1,2); 

  switch(attr)
  {
    case HEADING_MEASURED_VALUE:
      read_heading();                
      break;
                   
    default:
      errorPgm(string_unknown_attr);
      break;
  }  
}

void XBee::read_tilt_measurement_cluster(int i)
{
  int attr = _msgBuffer[i+1] << 8;
  attr = attr + _msgBuffer[i+2];
  attr = attr & 0xFFFF;
  //printHex(_msgBuffer+i+1,2); 

  read_tilt(attr);
    
}

void XBee::read_temp()
{
               
  printPgm(string_read_temperature);
  float a = getTemperature();
  //TODO:print("got temperature reading\r\n");
  int source_endpoint,destination_endpoint,datasize;
                
  //TODO: send response back
  //7e 00 xx 11 52 aa aa aa aa aa aa aa aa bb bb cc dd ee ee ff ff gg hh 00 01 00 00 ch
  source_endpoint = 1;
  destination_endpoint = 1;
  datasize = 9;
  byte data[datasize];
  memset(data,0,datasize);
  // data 18 01 01 00 00 00 29  09 f6
  data[0] = 0x18; // 00011000 0x18 general command = 00/ manufacturer specific = 0/ direction =1/ disable default rsp = 1 / reserver = 000
  data[1] = _transaction_sequence; // transaction sequence number = 01    
  data[2] = 0x01; // command identifier = 01 /read response
  
  // read attribute status record
  data[3] = 0x00; // attribute identifier = 00 00 / temperature
  data[4] = 0x00; // attribute identifier = 00 00 / temperature
  data[5] = 0x00; // status = 00 SUCCESS
  data[6] = 0x29;// attribute data type = 29 signed 16 bit integer
  if (a == FLT_MAX) { //temperature measurement invalid
    data[7] = 0x80;
    data[8] = 0x00;
  } else {
    int temp = a * 100;
    data[7] = (temp>>8) & 0xff; // attribute data 09f6 (25.5 x100 = 2550) 
    data[8] = (temp) & 0xff; // attribute data 09f6 (25.5 x100 = 2550) 
  }
  printPgm(string_send_response);
  send(_sourceAddress64, _sourceAddress16, source_endpoint, destination_endpoint, TEMP_MEASUREMENT_CLUSTERID, HA_PROFILEID, data, datasize);
  

}

void XBee::read_temp_min()
{
  //TODO:not implemented yet
  errorPgm(string_not_implemented);
}

void XBee::read_temp_max()
{
  //TODO:not implemented yet
  errorPgm(string_not_implemented);
}

void XBee::read_temp_tolerance() 
{
  //TODO:not implemented yet
  errorPgm(string_not_implemented);
} 

void XBee::send(byte addr64[], byte addr16[], char source_endpoint, char destination_endpoint, int clusterid, int profileid, byte data[], int datasize)
{
  //7e 00 xx 11 52 aa aa aa aa aa aa aa aa bb bb cc dd ee ee ff ff gg hh 00 01 00 00 ch
  int len = 1 + 1+ 8 + 2 + 1 + 1 + 2 +2 + 1 +1 + datasize; //   api 1 + frameid 1 + 64bit addr 8 + 16bit addr 2 + source endpoint 1 + dest endpoing 1+
                                                           // cluster id 2 + profile id 2 + broad cast radius 1 + options 1
  int buflen = 4+len; // 1(7e) + 2(len) + 1(checksum)
  byte buf[buflen];
  buf[0] = 0x7e;
  buf[1] = len & 0xFF00;
  buf[2] = len & 0x00FF;
  buf[3] = 0x11;
  buf[4] = ++_frameId;
  
  
  // copy 64bit address
  for(int i=0;i<8;i++) 
  {
    buf[5+i] = addr64[i];
  }
  
  // copy 16bit address
  for(int i=0;i<2;i++) 
  {
    buf[13+i] = addr16[i];
  }
  
  buf[15] = source_endpoint;
  buf[16] = destination_endpoint;
  
  // cluster id
  buf[17] = (clusterid >> 8) & 0xff;
  buf[18] = clusterid & 0xff;
  
  //profile id
  buf[19] = (profileid >> 8) & 0xff; 
  buf[20] = profileid & 0xff; 
  
  //broadcast radius
  buf[21] = 0;
  
  //options
  buf[22] = 0;
  
  for(int i=0;i<datasize;i++) 
  {
    buf[23+i] = data[i];
  }
  
  int checksuma = 0;
  for(int i =0; i < len;i++) 
  {
    checksuma = checksuma + buf[i+3];
  }
  byte checksum = 0xFF - (checksuma & 0xFF);
  buf[buflen-1] = checksum;
 
  printPgm(string_sending);
  printHex(buf,buflen);
  
  //serial output
  for(int i =0;i<buflen;i++) 
  {
    Serial.print(buf[i],BYTE);
  }
}



void XBee::read_illuminance_measurement_cluster(int i)
{
  int attr = _msgBuffer[i+1] << 8;
  attr = attr + _msgBuffer[i+2];
  attr = attr & 0xFFFF;
  printHex(_msgBuffer+i+1,2); 

  switch(attr)
  {
    case ILLUMINANCE_MEASURED_VALUE:
      read_illuminance();                
      break;
    case ILLUMINANCE_MIN_MEASURED_VALUE:
      read_illuminance_min();                
      break;
    case ILLUMINANCE_MAX_MEASURED_VALUE:
      read_illuminance_max();                
      break;
    case ILLUMINANCE_TOLERANCE:
      read_illuminance_tolerance();                
      break;               
    default:
      errorPgm(string_unknown_attr);
      break;
  }  
}

void XBee::read_illuminance() 
{
  //TODO:not implemented yet
  printPgm(string_read_illuminance);
  int illuminance = getIlluminance();
  //TODO:print("got temperature reading\r\n");
  int source_endpoint,destination_endpoint,datasize;
                
  //TODO: send response back
  //7e 00 xx 11 52 aa aa aa aa aa aa aa aa bb bb cc dd ee ee ff ff gg hh 00 01 00 00 ch
  source_endpoint = 1;
  destination_endpoint = 1;
  datasize = 9;
  byte data[datasize];
  memset(data,0,datasize);
  // data 18 01 01 00 00 00 29  09 f6
  data[0] = 0x18; // 00011000 0x18 general command = 00/ manufacturer specific = 0/ direction =1/ disable default rsp = 1 / reserver = 000
  data[1] = _transaction_sequence; // transaction sequence number = 01    
  data[2] = 0x01; // command identifier = 01 /read response
  
  // read attribute status record
  data[3] = 0x00; // attribute identifier = 00 00 / illuminance
  data[4] = 0x00; // attribute identifier = 00 00 / illuminance
  data[5] = 0x00; // status = 00 SUCCESS
  data[6] = 0x29;// attribute data type = 29 signed 16 bit integer
  if (illuminance == INT_MAX) { //temperature measurement invalid
    data[7] = 0xFF;
    data[8] = 0xFF;
  } else {
    //todo: change iluminance value
    data[7] = (illuminance>>8) & 0xff; // attribute data  
    data[8] = (illuminance) & 0xff; // attribute data  
  }
  printPgm(string_send_response);
  send(_sourceAddress64, _sourceAddress16, source_endpoint, destination_endpoint, TEMP_MEASUREMENT_CLUSTERID, HA_PROFILEID, data, datasize);
  
} 

void XBee::read_illuminance_min() 
{
  //TODO:not implemented yet
  errorPgm(string_not_implemented);
} 

void XBee::read_illuminance_max() 
{
  //TODO:not implemented yet
  errorPgm(string_not_implemented);
} 

void XBee::read_illuminance_tolerance() 
{
  //TODO:not implemented yet
  errorPgm(string_not_implemented);
} 

int XBee::getIlluminance()
{
    if (_zb->fgetIlluminance != NULL) return _zb->fgetIlluminance();
    return INT_MAX; //TODO:float max  
}

void XBee::read_heading()
{
               
  printPgm(string_read_heading);
  float a = getHeading();
  int source_endpoint,destination_endpoint,datasize;
                
  //TODO: send response back
  //7e 00 xx 11 52 aa aa aa aa aa aa aa aa bb bb cc dd ee ee ff ff gg hh 00 01 00 00 ch
  source_endpoint = 1;
  destination_endpoint = 1;
  datasize = 9;
  byte data[datasize];
  memset(data,0,datasize);
  // data 18 01 01 00 00 00 29  09 f6
  data[0] = 0x38; // 00111000 0x18 general command = 00/ manufacturer specific = 0/ direction =1/ disable default rsp = 1 / reserver = 000
  data[1] = _transaction_sequence; // transaction sequence number = 01    
  data[2] = 0x01; // command identifier = 01 /read response
  
  // read attribute status record
  data[3] = 0x00; // attribute identifier = 00 00 
  data[4] = 0x00; // attribute identifier = 00 00 
  data[5] = 0x00; // status = 00 SUCCESS
  data[6] = 0x29;// attribute data type = 29 signed 16 bit integer
  if (a == FLT_MAX) { //temperature measurement invalid
    data[7] = 0x80;
    data[8] = 0x00;
  } else {
    int heading = a * 100;
    data[7] = (heading>>8) & 0xff; // attribute data 09f6 (25.5 x100 = 2550) 25.5 degrees
    data[8] = (heading) & 0xff; // attribute data 09f6 (25.5 x100 = 2550) 
  }
  printPgm(string_send_response);
  send(_sourceAddress64, _sourceAddress16, source_endpoint, destination_endpoint, HEADING_MEASUREMENT_CLUSTERID, HA_PROFILEID, data, datasize);
  

}

void XBee::read_tilt(int i)
{
               
  float a = getTilt(i);
 // float a = getHeading();
  
  int source_endpoint,destination_endpoint,datasize;
                
  //TODO: send response back
  //7e 00 xx 11 52 aa aa aa aa aa aa aa aa bb bb cc dd ee ee ff ff gg hh 00 01 00 00 ch
  source_endpoint = 1;
  destination_endpoint = 1;
  datasize = 9;
  byte data[datasize];
  memset(data,0,datasize);
  // data 18 01 01 00 00 00 29  09 f6
  data[0] = 0x38; // 00111000 0x18 general command = 00/ manufacturer specific = 0/ direction =1/ disable default rsp = 1 / reserver = 000
  data[1] = _transaction_sequence; // transaction sequence number = 01    
  data[2] = 0x01; // command identifier = 01 /read response
  
  // read attribute status record
  data[3] = (i & 0xFF00) >> 8; // attribute identifier = 00 00 
  data[4] = i & 0xFF; // attribute identifier = 00 00 
  data[5] = 0x00; // status = 00 SUCCESS
  data[6] = 0x29;// attribute data type = 29 signed 16 bit integer
  if (a == FLT_MAX) { // measurement invalid
    data[7] = 0x80;
    data[8] = 0x00;
  } else {
    int phi = (int)a*100;
    
    data[7] = (phi>>8) & 0xff; // attribute data  
    data[8] = (phi) & 0xff; // attribute data 
  }
  printPgm(string_send_response);
  send(_sourceAddress64, _sourceAddress16, source_endpoint, destination_endpoint, TILT_MEASUREMENT_CLUSTERID, HA_PROFILEID, data, datasize);
  
}

