#ifndef XBee_h
#define XBee_h


#include "WConstants.h"
#include <avr/pgmspace.h>

#undef int()
#undef float()
struct zigbee { 
   float (*fgetTemperature)();
   int (*fgetIlluminance)();
   float (*fgetTilt)(int i);
   float (*fgetHeading)();
   void (*fprogress)();
   void (*ferror)(char *msg);
   void (*fprint)(char *msg);
   void (*fprintHex)(byte *msg,int size);
   void (*fmodemStatus)(int status);
   
};
//TODO: write an init functio for zigbee struct to zeroe all members


enum state { WAITING_FOR_NEW_PACKET,READING_LEN1,
             READING_LEN2,READING_API_ID_VALUE,
             READING_FRAMEID,READING_CMD_DATA,READING_CHECKSUM };


#define MODEM_STATUS 0x8a
#define AT_CMD 0x08
#define AT_CMD_Q_PARAM_VALUE 0x09
#define AT_CMD_RES 0x88
#define AT_CMD_REQ 0x17
#define REMOTE_CMD_RES 0x97
#define ZB_TX_REQ 0x10
#define EXPLICIT_ADDR_ZB_CMD 0x11
#define ZB_TX_STATUS 0x8B
#define ZB_RX_PACKET 0x90
#define EXPLICIT_RX_IND 0x91
#define ZB_IO_DATA_SAMPLE_RX_IND 0x92
#define XB_SENSOR_READ_IND 0x94
#define NODE_ID_IND 0x95

//zigbee standard values
#define HA_PROFILEID 0x0104
// temp
#define TEMP_MEASUREMENT_CLUSTERID 0x0402
#define TEMP_MEASURED_VALUE 0x0000
#define TEMP_MIN_MEASURED_VALUE 0x0001
#define TEMP_MAX_MEASURED_VALUE 0x0002
#define TEMP_TOLERANCE 0x0003
//illuminance
#define ILLUMINANCE_MEASUREMENT_CLUSTERID 0x0400
#define ILLUMINANCE_MEASURED_VALUE 0x0000
#define ILLUMINANCE_MIN_MEASURED_VALUE 0x0001
#define ILLUMINANCE_MAX_MEASURED_VALUE 0x0002
#define ILLUMINANCE_TOLERANCE 0x0003
#define ILLUMINANCE_LIGHT_SENSOR_TYPE 0x0004

// heading
#define HEADING_MEASUREMENT_CLUSTERID 0xFC00
#define HEADING_MEASURED_VALUE 0x0000

// heading
#define TILT_MEASUREMENT_CLUSTERID 0xFC01
#define TILT_PHI_MEASURED_VALUE 0x0000
#define TILT_RHO_MEASURED_VALUE 0x0001
#define TILT_THETA_MEASURED_VALUE 0x0002


//enumerated status values used in the ZCL
#define STATUS_SUCCESS 0x00
#define STATUS_FAILURE 0x01
#define STATUS_MALFORMED_COMMAND 0x80
#define STATUS_UNSUP_CLUSTER_COMMAND 0x81
#define STATUS_UNSUP_GENERAL_COMMAND 0x82
#define STATUS_UNSUP_MANUF_CLUSTER_COMMAND 0x83
#define STATUS_UNSUP_MANUD_GENERAL_COMMAND 0x84
#define STATUS_INVALID_FIELD 0x85
#define STATUS_UNSUPPORTED_ATTRIBUTE 0x86
#define STATUS_INVALID_VALUE 0x87
#define STATUS_READ_ONLY  0x88
#define STATUS_INSUFFICIENT_SPACE 0x89
#define STATUS_DUPLICATE_EXISTS 0x8a
#define STATUS_NOT_FOUND 0x8b
#define STATUS_UNREPORTABLE_ATTRIBUTE 0x8c
#define STATUS_INVALID_DATA_TYPE 0x8d
#define STATUS_HARDWARE_FAILURE 0xc0
#define STATUS_SOFTWARE_FAILURE 0xc1
#define STATUS_CALIBRATION_ERROR 0xc2


//commands
#define READ_ATTRS 0x00
#define READ_ATTRS_RESPONSE 0x01
#define WRITE_ATTRS 0x02
#define WRITE_ATTRS_UNDIVIDED 0x03
#define WRITE_ATTRS_RES 0x04
#define WRITE_ATTRS_NO_RES 0x05
#define CONFIGURE_REPORTING 0x06
#define CONFIGURE_REPORTING_RESPONSE 0x07
#define READ_REPORTING_CONFIGURATION 0x08
#define READ_REPORTING_CONFIGURATION_RES 0x09
#define REPORT_ATTRS 0x0a
#define DEFAULT_RESPONSE 0x0b
#define DISCOVER_ATTRS 0x0C
#define DISCOVER_ATTRS_RES 0x0D


//modem status
#define HARDWARE_RST 0
#define WATCHDOG_TIMER_RST 1
#define ASSOCIATED 2
#define DISASSOCIATED 3
#define SYNC_LOST 4
#define COORDINATOR_REALIGNMENT 5
#define COORDINATOR_STARTED 6






void xbee_init(struct zigbee *zb);

class XBee
{
  public:
    XBee(struct zigbee *zb);
    void serve();
    void parse();
    
  private:
    struct zigbee *_zb;
    state _state;
    int _frameId;
    int _msgLength;
    int _clusterid;
    byte _sourceAddress64[8];
    byte _sourceAddress16[2];
    byte _apiIdValue;
    byte _transaction_sequence;
    byte _msgBuffer[25];
    int _msgBufferIndex;
    
    
    void error(char *msg);
    void errorPgm(PGM_P p);
    void print(char *msg);
    void printPgm(PGM_P str);
    void printHex(byte *msg,int size);
    void modemStatus(int status);
    void send(byte sourceAddress64[], byte sourceAddress16[], char source_endpoint, char destination_endpoint, int clusterid, int profileid, byte data[], int datasize);
    void zb_rx_packet();    
    void zb_explicit_rx_packet();
    void parse_rx_packet_payload(int i);
    void parse_general_command(int i);
    void parse_command(int i,boolean isManSpecific);    
    
    //temperature
    void read_temp_measurement_cluster(int i);
    void read_temp(); 
    void read_temp_min(); 
    void read_temp_max(); 
    void read_temp_tolerance(); 
    float getTemperature();
    
    //illuminance
    void read_illuminance_measurement_cluster(int i);
    void read_illuminance(); 
    void read_illuminance_min(); 
    void read_illuminance_max(); 
    void read_illuminance_tolerance(); 
    int getIlluminance();
    
    
    //heading
    void read_heading_measurement_cluster(int i);
    void read_heading(); 
    float getHeading();

    //tilt
    void read_tilt_measurement_cluster(int i);
    void read_tilt(int i); 
    float getTilt(int i);
    
    void sendDefaultResponseIfNeeded(int i);
    
};

#endif
